KGroup		group	IPF_Kernel,CSEQ_HNDLR_SEG
		locals
IPF_Kernel	segment	para	public
		assume	cs:IPF_Kernel
Start:;
		jmp	Init
;*************************************************************************
;
;	IPF kernel.  Resident part of the IPF
;	Release 1.02     (c) Max B. Motovilov
;
;	First written in 1991.
;	Minor changes in 1992.
;	Bugs fixed & comments added in January 1993.
;
;	...Max...
;
;*************************************************************************

ASCII_PAGE	equ	1
GRAPH_PAGE	equ	2
CYR_PAGE	equ	3

PSP_PARS	equ	10h

CURRENT_VERSION	equ	0102h

PRINTER_OK	equ	10h
IO_ERR_BITS	equ	09h


INT_17_Vector	dd	?
Port		dw	0
Handler_Segment	dw	?
Font_Page	db	?
IPF_OFF		db	0
IPF_Reset	db	1

;************************************************************************
;
;	INT 17h processing entry
;
;	Processes all standard requests plus:
;
;		AH =	0FFh		-- IPF on/off switching
;
;	Input:
;		AL =	0		-- Turn IPF off
;		AL =	1		-- Turn IPF on
;
;	Output:
;		AL =    0FFh		-- Use to check IPF presence
;		AH =    Printer status
;
;************************************************************************

INT_17:;
		sti
		cmp	ah,0ffh			;	If it is IPF's
		jne	@@1                 	;	own function
		or	al,al
		jz	@@2	                ;	On ?
		mov	cs:IPF_OFF,0            ;	Turn ON
		mov	cs:IPF_Reset,1          ;	Printer reset requied
		jmp	short Mainline          ;	Do reset & exit
@@2:		mov	cs:IPF_OFF,1            ;	Else: Turn OFF
		mov	cs:IPF_Reset,1          ;	- * -
		jmp	short Mainline          ;
@@1:		cmp	dx,cs:Port		;	Our LPT ?
		je	@@4
@@3:		cli        			;	Jump to BIOS 17h
		jmp	cs:dword ptr INT_17_Vector
@@4:		or	ah,ah                   ;
		jnz	@@3                   	;	Process only AH=0
		cmp	cs:IPF_OFF,0            ;       If we're OFF,
		jne	@@3			;	do nothing

;**************************************************************************
;
;	Here we get only if AH=FF or it is "print character" request
;	to our LPT
;
;**************************************************************************

Mainline:;
		push	bx
		push	cx
		push	dx
		push	si
		push	ds			;	Save regs
		mov	ds,cs:Handler_Segment
		assume	ds:CSEQ_HNDLR_SEG       ;	Setup DS:

		cmp	cs:IPF_Reset,0          ;  	Are we performing
		je	Check_Controls          ;	printer reset?
		push	ax			;	Save user's AX
		mov	ah,2                    ;
		mov	dx,cs:Port              ;	Check status:
		call	Do_Int17                ;	INT 17, AH=2
		jnc	@@ok                    ;	Printer error?
		pop	dx			;	Restore old AX into DX
		cmp	dh,0ffh                 ;
		jne	@@exit                  ;     	Our function?
		mov	al,dh                   ;	AL=0FFh (AH still
		jmp	short @@exit            ;	contains prn. status)

@@ok:		mov	cs:IPF_Reset,0          ;	Reset complete
		call	far ptr RESET_Caller    ;	Do it!
		pop	dx			;	DX= old AX
		cmp	dh,0ffh                 ;
		jne	Check_Controls          ;	Our function?
		mov	al,dh                   ;	- * - (see prn. err.)
@@exit:		jmp	Exit_Mainline           ;


Call_Ptr        dd      ?			;	Address for indirect
						;	call

RESET_Caller:   mov     word ptr cs:[Call_Ptr+2],ds
		push	RESET_Entry             ; 	Indirect FAR call
                pop     word ptr cs:Call_Ptr    ;
                jmp     cs:[Call_Ptr]           ;
;                                               ;	This simpler method
;                push    ds                     ;       shows some strange
;                push    RESET_Entry            ;       effects on 386s...
;                retf                           ;       Why???
;                                               ;

;**************************************************************************
;
;	Control sequences processing.
;	We get there only in case of "print character" (AH=0)
;
;**************************************************************************

Check_Controls:	mov	dx,CONTROL_Entry	;	If we still have
		or	dx,dx                   ;       a control sequence
		jz	No_CTRL_Or_GROUP        ;       to process:

		cmp	dx,End_Of_Table         ;	If DX points to a table,
		ja	Call_CONTROL            ;       not driver's entry:

		mov	si,dx			;	Setup SI and AH for
		mov	ah,al                   ;       scanning the table
		cld                             ;

G_Loop:		lodsb				;	Get table entry
		cmp	al,0ffh                 ;
		je	G_List                  ;       Isn't it a list?
		or	al,al                   ;
		jz	G_End                   ;       Not an end of table?
		cmp	al,ah                   ;
		je	G_Found                 ;       Don't chars match?
		jmp	short G_Skip            ;	It's a mess, skip it

;	List processing. Applied for similarly treated sequences.

G_List:		lodsb                           ;	Get list entry
		or	al,al			;
		jz	GL_End                  ;	Not an end of list?
		cmp	al,ah                   ;	Match?
		jne	G_List                  ;       Scan until match

GL_Skip:	lodsb                           ;	Skip the trailing
		or	al,al                   ;       entries until the
		jnz	GL_Skip                 ;       end of list

		jmp	short G_Found		;	OK, we've got it!
GL_End:;

;	End of list processing part. Here we go...

G_Skip:		add	si,2                    ;	Skip entry address
		jmp	short G_Loop            ;	Continue scanning

;	Here we get if a character doesn't belong to any known control
;	sequence from the group we process.

G_End:		mov	si,BUFFER_Address	;       Output the beginning
		mov	[si+1],ah               ;       of this group together
		mov	cx,2                    ;	with the current char
		call	Printer_Output_Routine  ;       to printer

		xor	ax,ax			;       Clear the address
		mov	CONTROL_Entry,ax        ;       of curr. sequence

		mov	ah,PRINTER_OK		;	Simulate normal exit
		jmp	Exit_Mainline           ;       on INT 17, AH=0

;	Here we get if we happen to find a matching entry in the table

G_Found:	mov	dx,[si]			;	DX=entry address
		mov	al,ah             	;	AL=last character

;***************************************************************************
;
;	Calling the driver's entry.
;	By now, DX=entry address, AL=last character
;
;***************************************************************************

Call_CONTROL:	mov	si,BUFFER_Address	;	SI=saving buffer
		mov	bx,Fill_Pointer		;	BX=fill pointer
		mov	[bx],al			;	Save last character
		inc	bx                      ;	& move BX
		mov	cx,bx                   ;
		sub	cx,si                   ;	CX=fill count
		call	far ptr CTRL_Caller     ;	Call entry

		mov	Fill_Pointer,bx         ;	Save fill pointer &
		mov	CONTROL_Entry,dx        ;       chain link

		mov	ah,PRINTER_OK		;	Simulate normal exit
		jmp	Exit_Mainline           ;

CTRL_Caller:					;	Indirect caller
                mov     word ptr cs:[Call_Ptr+2],ds
                mov     word ptr cs:[Call_Ptr],dx
                jmp     cs:[Call_Ptr]

;                push    ds                     ;	Same mess with 386...
;                push    dx                     ;
;                retf                           ;

;***************************************************************************
;
;	Character processing. We get here in the most of the cases.
;	There's no current control sequence.
;
;***************************************************************************

No_CTRL_Or_GROUP:;
		cmp	al,' '			; 	Could this character
		ja	Printed_Character       ;       begin a sequence?

		lea	si,CSEQ_Table		;	Setup SI and AL
		mov	ah,al                   ;	for scanning the
		cld                             ;       table

C_Loop:		lodsb				;	Get table entry
		or	al,al                   ;
		jz	C_End                   ;	Not an end of table?
		cmp	al,0ffh                 ;
		je	C_Group                 ;	Not a group?
		cmp	ah,al                   ;
		je	C_Found                 ;	Doesn't match?

		add	si,2                    ;	Then skip it and
		jmp	short C_Loop            ;       continue scanning

;	We get here if we encounter a group of control sequences, beginning
;	from the same character (e.g. ESC-sequences).

C_Group:	lodsb				;	Get the leading char
		cmp	ah,al                   ;
		je	CG_Found                ;       Doesn't match?

		lodsb				;	Get the count
		mov	dl,al			;
		xor	dh,dh                   ;       Move SI by count
		add	si,dx                   ;       bytes
		jmp	short C_Loop            ;       Continue scanning

;	We get here if the initial character of the group matches the
;	character to print.

CG_Found:	inc	si			;	Skip count
		mov	CONTROL_Entry,si        ;	Setup chain
		mov	bx,BUFFER_Address       ;	Set BX to buffer
		mov	[bx],ah                 ;
		inc	bx                      ;	Save character
		mov	Fill_Pointer,bx		;	Setup fill pointer
		mov	ah,PRINTER_OK		;	Simulate normal		
		jmp	Exit_Mainline           ;       exit

;	We get here if table entry matches our character

C_Found:	mov	dx,[si]			;	DX = entry point
		jmp	Call_CONTROL		;	Call handler

;	We get here if none of the entries match

C_End:		mov	al,ah             	;	Restore AL
		xor	ah,ah			;	Clear AH
		jmp	Exit_And_Do		;	Proceed with INT 17

;***************************************************************************
;
;	Printable character processing.
;	We get here in case of AH=0, AL>SPACE and no pending control
;	sequences.
;
;***************************************************************************

Printed_Character:;
		mov	dh,cs:Font_Page		;	Get current font page
		xor	dl,dl			;	Set decrement to 0
		cmp	al,'0'			;
		jae	@@2	                ;	AL = symbol ?

		mov	ah,dh                   ;	Page doesn't matter
		jmp	short @@1		;

@@2:		cmp	al,80h			;
		jae	@@3	                ;	AL = ASCII char ? 

		mov	ah,ASCII_PAGE           ;	Request ASCII page
		jmp	short @@1		;

@@3:		cmp	al,0b0h			;	
		jae	@@4	                ;       AL = CYRILLIC char ?

		mov	ah,CYR_PAGE		;	Request CYR page
		mov	dl,080h-'0'             ;       Set proper decrement
		jmp	short @@1	        ;

@@4:		cmp	al,0e0h			;
		jae	@@5                     ;	AL = GRAPHIC char ?

		mov	ah,GRAPH_PAGE		; 	Request GRAPH page  
		jmp	short @@1               ;

@@5:		mov	ah,CYR_PAGE		;	Request CYR page
		mov	dl,0e0h-'`'             ;       Set proper decrement
						;	(for 'r' -- 'ja')

@@1:		call	far ptr PRINT_Caller	;	Call handler's entry
		mov	cs:Font_Page,ah		;	Save font page
		xor	ah,ah			;	Restore AH (=0)
		jmp	Exit_And_Do		;	Proceed with INT 17

PRINT_Caller:   mov     word ptr cs:[Call_Ptr+2],ds
						;	Indirect caller
		push	PRINT_Entry
                pop     word ptr cs:Call_Ptr
                jmp     cs:[Call_Ptr]

;                push    ds
;                push    PRINT_Entry
;                retf

;**************************************************************************
;
;	Exit without actually calling INT 17
;
;**************************************************************************

Exit_Mainline:;
		pop	ds
		pop	si
		pop	dx
		pop	cx
		pop	bx
		iret

;*************************************************************************
;
;	Normally proceeding with INT 17
;
;*************************************************************************

Exit_And_Do:;
		pop	ds
		pop	si
		pop	dx
		pop	cx
		pop	bx
		jmp	cs:INT_17_Vector

;*************************************************************************
;
;	Simulating INT 17 for internal calls
;
;*************************************************************************

Do_Int17	proc	near

		pushf				;
		cli                             ;      	INT 17 simulation
		call	cs:INT_17_Vector        ;

		test	ah,IO_ERR_BITS		;	
		jnz	@@io_error		;	OK ?

		clc				;
		ret                             ;	Exit, CF=0

@@io_error:	stc				;
		ret                             ;	Exit, CF=1

		endp

;*************************************************************************
;
;	Output routine to be called by handler's code
;
;	Expects:
;		SI to point to start of the buffer
;		CX to contain byte counter
;
;	Doesn't destroy any of the registers.
;
;*************************************************************************

Printer_Output_Routine	proc	far
		push	dx
		push	si
		push	ax
		push	cx			;	Save all regs

		mov	dx,cs:Port		;	Get LPT# into DX

@@loop:		lodsb				;	Get byte

@@out:		xor	ah,ah                   ;
		call	Do_Int17	        ;	Try to print it...
		jc	@@out			;	...until it worked

		loop	@@loop			;	Proceed with next one

		pop	cx
		pop	ax
		pop	si
		pop	dx			;	Restore regs
		ret
		endp

;*************************************************************************
;
;	A short code which transfers the handler into a proper
;	position and TSRs
;
;*************************************************************************

Do_TSR:		rep 	movsb
		int	21h

End_Of_Kernel:;

;************************************************************************
;
;	IPF kernel is kept resident up to here
;
;************************************************************************

;	INITIALIZATION CODE FOLLOWS

;************************************************************************
;
;	Init-time output routine
;
;************************************************************************

Init_Do_Int17	proc	near				
		int	17h
Check_Int17	label	proc
		test	ah,9h
		jnz	@@io_error
		clc
		ret
@@io_error:	stc
		ret
		endp


SKIP		macro	byte,id
		ifdif	<byte>,<al>
			mov	al,byte
		endif
LL_&id:		scasb
		jne	LL2_&id
		loop	LL_&id
		jmp	short End_Of_Args
LL2_&id:	dec	di
		endm

FIND		macro	byte,id
		ifdif	<byte>,<al>
			mov	al,byte
		endif
		repne	scasb
		jcxz	LF_&id
		dec	di
		inc	cx
LF_&id:;
		endm

MESSAGE		macro	msg
		lea	dx,msg
		mov	ah,09h
		int	21h
		endm

Driver_Name	dw	0
Font_Name	dw	0
Loaded_Size	dw	0
;
Ret_Code	dw	4c00h
;
MSG_Wrong_Option	db	"Wrong option '/"
Msg_Opt_Byte		db 	" "
			db	"' given in command line",0ah,0dh,"$"
MSG_Hello	label	byte
	db	"Intelligent Printer Font Handling.",0ah,0dh
	db	"By Max.B.Motovilov.",0ah,0dh
	db	"IPF Kernel : Release 1.02 from Jan. 1993",0ah,0dh,0ah,0dh,"$"
MSG_Help	label	byte
	db	"Usage : IPF [/n] Driver Font",0ah,0dh
	db	"Or    : IPF  /f  Font",0ah,0dh,0ah,0dh
	db	"Driver and Font are full pathnames.",0ah,0dh,0ah,0dh
	db	"/n  --  Set IPF for LPTn ( n can be 0,1 or 2 ).",0ah,0dh
	db	"/f  --  Do not install IPF, load font only.",0ah,0dh
	db      "        IPF should be installed earlier!",0ah,0dh,"$"
;
MSG_Already_Installed	label	byte
	db	"IPF is already installed -- cannot load new driver",0ah,0dh,"$"
;
MSG_No_IPF	label	byte
	db	"IPF not installed. Please specify driver.",0ah,0dh,"$"
;
MSG_Cant_Load_Font	label	byte
	db	"Printer is offline or not ready. Cannot load font.",0ah,0dh
	db	"You should use 'IPF /f Font' after turning on the printer",0ah,0dh,"$"
;
MSG_No_Driver	label	byte
	db	"Cannot find driver file.",0ah,0dh,"$"
;
MSG_No_Font	label	byte
	db	"Cannot find font file.",0ah,0dh,"$"
;
MSG_IO_Error	label	byte
	db	"I/O error occured while reading file.",0ah,0dh,"$"
;
MSG_Long_Driver	label	byte
	db	"Driver file too long."
;
MSG_Long_Font	label	byte
	db	"Font file too long."
;
MSG_Wrong_Version	label	byte
	db	"Driver requires IPF kernel release "
Ver_High_Byte	db	"0"
		db	"."
Ver_Low_Byte	db	"0"
	db	" or higher.",0ah,0dh,"$"
Init:;
		assume	ds:IPF_Kernel
		push	cs
		pop	ds
		MESSAGE	MSG_Hello
;
;**************  Parsing command line  ***************
;
		xor	ch,ch
		mov	cl,es:[80h]
		mov	di,81h
		cld
;
		or	cx,cx
		jnz	T_099
		jmp	End_Of_Args
T_099:		SKIP	' ',1
		cmp	es:byte ptr [di],'/'
		jne	Get_Driver_Name
		scasb
		dec	cx
		cmp	es:byte ptr [di],'f'
		jne	T_100
		scasb
		dec	cx
		jcxz	Show_Help
		jmp	short Get_Font_Name
T_100:		cmp	es:byte ptr [di],'0'
		jb	Wrong_Option
		cmp	es:byte ptr [di],'2'
		ja	Wrong_Option
		mov	al,es:[di]
		xor	ah,ah
		sub	al,'0'
		mov	Port,ax
		scasb
		dec	cx
		jcxz	Show_Help
		SKIP	' ',2
Get_Driver_Name:;
		mov	Driver_Name,di
		FIND	' ',1
		mov	es:byte ptr [di],0
		jcxz	End_Of_Args
		scasb
		dec	cx
Get_Font_Name:;
		SKIP	' ',3
		mov	Font_Name,di
		FIND	' ',2
		mov	es:byte ptr [di],0
		jmp	End_Of_Args
Wrong_Option:;
		mov	al,es:[di]
		mov	Msg_Opt_Byte,al
		MESSAGE	MSG_Wrong_Option
		mov	ax,4c01h
		int	21h
End_Of_Args:;
;
;**************  Loading driver & font  ***************
;
		cmp	Driver_Name,0
		jne	T_120
		cmp	Font_Name,0
		jne	T_120
Show_Help:	MESSAGE	MSG_Help
		mov	ax,4c01h
		int	21h
T_120:;
		mov	ax,0FF00h
		int	17h
		push	ax
		cmp	Driver_Name,0
		je	T_130
		cmp	al,0ffh
		je	@@1
		call	Load_Driver
		jmp	short @@2
@@1:		MESSAGE	MSG_Already_Installed
@@2:		pop     ax
		push	ax
		cmp	al,0ffh
		je	T_130a
		mov	ah,2
		mov	dx,Port
		call	Init_Do_Int17
		jmp	short T_130a
T_130:		cmp	al,0ffh
		je	T_130a
		MESSAGE	MSG_No_IPF
		mov	ax,4c01h
		int	21h
T_130a:		call	Check_Int17
		jnc	T_131
		MESSAGE	MSG_Cant_Load_Font
		mov	Ret_Code,4c01h
		jmp	short T_132
T_131:		cmp	Font_Name,0
		je	T_132
		call	Load_Font
T_132:		pop	ax
		cmp	al,0ffh
		jne	T_133
		mov	ax,0ff01h
		int	17h
T_133:;
;
;**************  Final preparations & TSR  ***************
;
		cmp	Loaded_Size,0
		jne	T_140
		mov	ax,Ret_Code
		int	21h
T_140:;
		mov	ax,03517h
		int	21h
		mov	word ptr INT_17_Vector,bx
		mov	word ptr INT_17_Vector+2,es
		lea	dx,INT_17
		mov	ax,02517h
		int	21h
		mov	ax,ds
		lea	dx,End_Of_Kernel
		sar	dx,1
		sar	dx,1
		sar	dx,1
		sar	dx,1
		inc	dx
		add	dx,ax
		mov	es,dx
		mov	Handler_Segment,dx
		mov	ax,seg CSEQ_HNDLR_SEG
		mov	ds,ax
		lea	dx,Printer_Output_Routine
		assume	ds:CSEQ_HNDLR_SEG
		mov	word ptr Print_Routine,dx
		xor	si,si
		mov	di,si
		mov	cx,cs:Loaded_Size

		mov	dx,cx
		sar	dx,1
		sar	dx,1
		sar	dx,1
		sar	dx,1
		inc	dx		;	DX = driver size

		mov	ax, es
		add	dx, ax

		mov	ax, cs
		sub	dx, ax        	;	DX = driver + IPF

		add	dx, PSP_PARS	;	DX = driver + IPF + PSP

		mov	ax,3100h
		jmp	Do_TSR
;
;**************  Loading driver  ***************
;
Load_Driver	proc	near
		mov	dx,Driver_Name
		push	ds
		push	es
		pop	ds
		mov	ax,3d00h
		int	21h
		pop	ds
		jnc	T_160
		MESSAGE	MSG_No_Driver
		mov	ax,4c01h
		int	21h
T_160:		mov	bx,ax
		mov	ax,4202h
		xor	cx,cx
		mov	dx,cx
		int	21h
		jnc	T_165
		jmp	IO_Error
T_165:		or	dx,dx
		jz	T_161
		MESSAGE	MSG_Long_Driver
		mov	ax,4c01h
		int	21h
T_161:		mov	Loaded_Size,ax
		mov	ax,4200h
		int	21h
		mov	cx,Loaded_Size
		mov	ah,3fh
		push	ds
;
		assume	ds:CSEQ_HNDLR_SEG
		mov	dx,seg CSEQ_HNDLR_SEG
		mov	ds,dx
;
		xor	dx,dx
		int	21h
		jnc	T_168
		jmp	IO_Error
T_168:		mov	ah,3eh
		int	21h
		mov	ax,INIT_Entry
		mov	cs:Loaded_Size,ax
		mov	dx,cs
		mov	word ptr Print_Routine+2,dx
		lea	dx,Init_Print
		mov	word ptr Print_Routine,dx
		call	far ptr INIT_Caller
		push	ax
	        mov	ah,2
		mov	dx,cs:Port
		call	Init_Do_Int17
		jc	@@1
		call	far ptr RESET_Caller
@@1:		pop	ax
		cmp	ax,0ffffh
		jne	T_163
		mov	ax,4c01h
		int	21h
T_163:		cmp	ax,CURRENT_VERSION
		jbe	T_162
		assume	ds:IPF_Kernel
		pop	ds
		add	Ver_High_Byte,ah
		add	Ver_Low_Byte,al
		MESSAGE	MSG_Wrong_Version
		mov	ax,4c01h
		int	21h
		assume	ds:CSEQ_HNDLR_SEG
T_162:;
		assume	ds:IPF_Kernel
		pop	ds
		ret
		endp
;
INIT_Caller:    mov     word ptr cs:[Call_Ptr+2],ds
                mov     word ptr cs:[Call_Ptr],ax
                jmp     cs:[Call_Ptr]
;
;                push    ds
;                push    ax
;                retf
;
Init_Print	proc	far
		push	si
		push	ax
		push	dx
		push	cx
		mov	dx,cs:Port
		cld
IP_Loop:	lodsb
IP2:		xor	ah,ah
		call	Init_Do_Int17
		jc	IP2
		loop	IP_Loop
		pop	cx
		pop	dx
		pop	ax
		pop	si
		ret
		endp
;
;**************  Loading font  **************
;
Load_Font	proc	near
		mov	dx,Font_Name
		push	ds
		push	es
		pop	ds
		mov	ax,3d00h
		int	21h
		pop	ds
		jnc	T_170
		MESSAGE	MSG_No_Font
		mov	ax,4c01h
		int	21h
T_170:		mov	bx,ax
		mov	ax,4202h
		xor	cx,cx
		mov	dx,cx
		int	21h
		jnc	T_175
		jmp	IO_Error
T_175:		or	dx,dx
		jz	T_171
		MESSAGE	MSG_Long_Font
		mov	ax,4c01h
		int	21h
T_171:		push	ax
		mov	ax,4200h
		int	21h
		pop	cx
		mov	ah,3fh
		push	ds
;
		mov	dx,Loaded_Size
		sar	dx,1
		sar	dx,1
		sar	dx,1
		sar	dx,1
		inc	dx
		add	dx,seg CSEQ_HNDLR_SEG
		mov	ds,dx
;
		xor	dx,dx
		int	21h
		jc	IO_Error
		mov	ah,3eh
		int	21h
		mov	dx,cs:Port
		xor	si,si
		cld
		lodsw
		mov	cx,ax
		call	Init_Print
		pop	ds
		ret
		endp
;
;**********************************************************
;
IO_Error:	push	cs
		pop	ds
		MESSAGE	MSG_IO_Error
		mov	ax,4c01h
		int	21h
		ends
CSEQ_HNDLR_SEG	segment	para public
INIT_Entry	dw	?
Print_Routine	dd	?
RESET_Entry 	dw	?
PRINT_Entry	dw	?
CONTROL_Entry	dw	?
End_Of_Table	dw	?
BUFFER_Address	dw	?
Fill_Pointer	dw	?
CSEQ_Table	db	?
		ends
	end 	Start
